#include-once
#region ;**** Include My Classes						****
#include "AutoItobject.au3"
#endregion ;**** Include My Classes						****
Global $Class_oComError, $Class_FlagDebug = False

Func __Item__(ByRef $Data, $NextEl = 0)
	Local $oItem = _AutoItObject_Create()
	_AutoItObject_AddProperty($oItem, "Data", $ELSCOPE_PUBLIC, $Data)
	_AutoItObject_AddProperty($oItem, "Next", $ELSCOPE_PUBLIC, $NextEl)
	Return $oItem
EndFunc   ;==>__Item__

Func _oCollection($Name = "")
	Local $oCollection = _AutoItObject_Create()
	_AutoItObject_AddProperty($oCollection, "First")
	_AutoItObject_AddProperty($oCollection, "Last")
	_AutoItObject_AddProperty($oCollection, "Size", -1)
	_AutoItObject_AddProperty($oCollection, "Name", $ELSCOPE_READONLY, $Name)

	_AutoItObject_AddMethod($oCollection, "Count", "_oCollection_Count")
	_AutoItObject_AddMethod($oCollection, "Debug", "_oCollection__Debug")
	_AutoItObject_AddMethod($oCollection, "Push", "_oCollection_Push")
	_AutoItObject_AddMethod($oCollection, "Insert", "_oCollection_Insert")
	_AutoItObject_AddMethod($oCollection, "Items", "_oCollection_Items")
	_AutoItObject_AddMethod($oCollection, "Remove", "_oCollection_Remove")
	_AutoItObject_AddMethod($oCollection, "Remove_All", "_oCollection_Remove_All")

	; Add enum
	_AutoItObject_AddEnum($oCollection, "_oCollection_EnumNext", "_oCollection_EnumReset")
;~ 	_Debug2Console("Created collection: " & $Name)
	Return $oCollection
EndFunc   ;==>_oCollection
Func _oCollection__Debug($oCollection, $bReturn = False, $ScriptName = @ScriptName, $LineNumber = @ScriptLineNumber)
	Local $iIndex = 0, $Return = "Items Count: " & $oCollection.Count & @LF

	For $Item In $oCollection
		If IsObj($Item) Then
			$Return &= "--Index: " & $iIndex & ", " & $Item.Debug(True, $ScriptName, $LineNumber)
		ElseIf IsArray($Item) Then
			$Return &= "--Index: " & $iIndex & ", Array[]"
		Else
			$Return &= "--Index: " & $iIndex & ", Value: " & $Item
		EndIf
		$iIndex += 1
	Next

	If $bReturn Then Return $Return
	If $Class_FlagDebug Then ConsoleWrite("_oCollection__Debug: " & $oCollection.Name & @CRLF & $Return)
	ConsoleWrite("_oCollection__Debug: " & $oCollection.Name & @CRLF & $Return)
EndFunc   ;==>_oCollection__Debug
Func _oCollection_Count($oCollection)
	If $oCollection.Size = -1 Then Return SetError(1, 0, -1)
	Return Int($oCollection.Size)
EndFunc   ;==>_oCollection_Count
Func _oCollection_EnumReset(ByRef $oCollection, ByRef $iterator)
	#forceref $oCollection
	$iterator = 0
EndFunc   ;==>_oCollection_EnumReset
Func _oCollection_EnumNext(ByRef $oCollection, ByRef $oIiterator)
	If $oCollection.Count = 0 Then Return SetError(1, 0, 0)
	If Not IsObj($oIiterator) Then
		$oIiterator = $oCollection.First
		Return $oIiterator.Data
	EndIf
	If Not IsObj($oIiterator.Next) Then Return SetError(1, 0, 0)
	$oIiterator = $oIiterator.Next
	Return $oIiterator.Data
EndFunc   ;==>_oCollection_EnumNext
Func _oCollection_Insert($oCollection, $iIndex, ByRef $NewData)
	Local $iCount = $oCollection.Count
	If $iCount = 0 Then
		$oCollection.First = __Item__($NewData)
		$oCollection.Last = $oCollection.First
	ElseIf $iIndex >= $iCount Then
		Local $oLast = $oCollection.Last
		$oLast.Next = __Item__($NewData)
		$oCollection.Last = $oLast.Next
	ElseIf $iIndex = 0 Then
		Local $oFirst = $oCollection.First
		$oCollection.First = __Item__($NewData, $oFirst)
	Else
		Local $i = 0, $oCurrent = $oCollection.First, $oPrevious = 0
		Do
			If $i = $iIndex Then
				$oPrevious.Next = __Item__($NewData, $oCurrent)
				ExitLoop
			EndIf
			$i += 1
			$oPrevious = $oCurrent
			$oCurrent = $oCurrent.Next
		Until $oCurrent = 0
	EndIf
	$oCollection.Size = $iCount + 1
EndFunc   ;==>_oCollection_Insert
Func _oCollection_Items($oCollection, $iIndex)
	Local $i = 0
	For $oItem In $oCollection
		If $i = $iIndex Then Return $oItem
		$i += 1
	Next
	Return SetError(1, 0, 0)
EndFunc   ;==>_oCollection_Items
Func _oCollection_Push($oCollection, ByRef $NewData)
	Local $iCount = $oCollection.Count
	Local $oLast = $oCollection.Last
	If $iCount = 0 Then
		$oCollection.First = __Item__($NewData)
		$oCollection.Last = $oCollection.First
	Else
		$oLast.Next = __Item__($NewData)
		$oCollection.Last = $oLast.Next
	EndIf
	$oCollection.Size = $iCount + 1
EndFunc   ;==>_oCollection_Push
Func _oCollection_Remove($oCollection, $iIndex)
	If $oCollection.Count = 0 Then Return SetError(1, 0, 0)
	Local $i = 0, $oCurrent = $oCollection.First, $oPrevious = 0
	Do
		If $i = $iIndex Then
			If $oCollection.Count = 1 Then
				; very Last element
				$oCollection.First = 0
				$oCollection.Last = 0
			ElseIf $i = 0 Then
				; First element
				$oCollection.First = $oCurrent.Next
			Else
				If $i = $oCollection.Count - 1 Then $oCollection.Last = $oPrevious ; Last element
				$oPrevious.Next = $oCurrent.Next
			EndIf
			$oCollection.Size = $oCollection.Count - 1
			Return
		EndIf
		$i += 1
		$oPrevious = $oCurrent
		$oCurrent = $oCurrent.Next
	Until $oCurrent = 0
	Return SetError(2, 0, 0)
EndFunc   ;==>_oCollection_Remove
Func _oCollection_Remove_All($oCollection)
	If $oCollection.Count = 0 Then Return SetError(1, 0, 0)
	Local $i = 0, $oCurrent = $oCollection.First, $oPrevious = 0
	Do
		If $oCollection.Count = 1 Then
			; very Last element
			$oCollection.First = 0
			$oCollection.Last = 0
		ElseIf $i = 0 Then
			; First element
			$oCollection.First = $oCurrent.Next
		Else
			If $i = $oCollection.Count - 1 Then $oCollection.Last = $oPrevious ; Last element
			$oPrevious.Next = $oCurrent.Next
		EndIf
		$oCollection.Size = $oCollection.Count - 1
		$i += 1
		$oPrevious = $oCurrent
		$oCurrent = $oCurrent.Next
	Until $oCurrent = 0
	Return SetError(2, 0, 0)
EndFunc   ;==>_oCollection_Remove_All